from __future__ import annotations

import asyncio
import functools
import inspect
import io
import itertools
import json
import logging
import numbers
import re
import time
import traceback
import uuid
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Annotated,
    Any,
    Awaitable,
    Callable,
    Concatenate,
    Generic,
    Hashable,
    Mapping,
    MutableMapping,
    Optional,
    ParamSpec,
    Protocol,
    Tuple,
    TypeAlias,
    TypeVar,
    Union,
)

import sqlalchemy as sa
import trafaret as t
import yaml
from aiohttp import web
from aiohttp.typedefs import Handler
from pydantic import Field, TypeAdapter, ValidationError

from ai.backend.common.api_handlers import BaseRequestModel, BaseResponseModel
from ai.backend.common.json import load_json
from ai.backend.common.types import AccessKey
from ai.backend.logging import BraceStyleAdapter

from ..models import UserRole, users
from ..utils import (
    check_if_requester_is_eligible_to_act_as_target_access_key,
    check_if_requester_is_eligible_to_act_as_target_user_uuid,
)
from .exceptions import (
    DeprecatedAPI,
    GenericForbidden,
    InvalidAPIParameters,
    NotImplementedAPI,
)

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

_rx_sitepkg_path = re.compile(r"^.+/site-packages/")


def method_placeholder(orig_method):
    async def _handler(request):
        raise web.HTTPMethodNotAllowed(request.method, [orig_method])

    return _handler


async def get_access_key_scopes(
    request: web.Request, params: Optional[Any] = None
) -> Tuple[AccessKey, AccessKey]:
    if not request["is_authorized"]:
        raise GenericForbidden("Only authorized requests may have access key scopes.")
    root_ctx: RootContext = request.app["_root.context"]
    owner_access_key: Optional[AccessKey] = (params or {}).get("owner_access_key", None)
    if owner_access_key is None or owner_access_key == request["keypair"]["access_key"]:
        return request["keypair"]["access_key"], request["keypair"]["access_key"]
    async with root_ctx.db.begin_readonly() as conn:
        try:
            await check_if_requester_is_eligible_to_act_as_target_access_key(
                conn,
                request["user"]["role"],
                request["user"]["domain_name"],
                owner_access_key,
            )
            return request["keypair"]["access_key"], owner_access_key
        except ValueError as e:
            raise InvalidAPIParameters(str(e))
        except RuntimeError as e:
            raise GenericForbidden(str(e))


async def get_user_uuid_scopes(
    request: web.Request, params: Optional[Any] = None
) -> Tuple[uuid.UUID, uuid.UUID]:
    if not request["is_authorized"]:
        raise GenericForbidden("Only authorized requests may have access key scopes.")
    root_ctx: RootContext = request.app["_root.context"]
    owner_uuid: Optional[uuid.UUID] = (params or {}).get("owner_uuid", None)
    if owner_uuid is None or owner_uuid == request["user"]["uuid"]:
        return request["user"]["uuid"], request["user"]["uuid"]
    async with root_ctx.db.begin_readonly() as conn:
        try:
            await check_if_requester_is_eligible_to_act_as_target_user_uuid(
                conn,
                request["user"]["role"],
                request["user"]["domain_name"],
                owner_uuid,
            )
            return request["user"]["uuid"], owner_uuid
        except ValueError as e:
            raise InvalidAPIParameters(str(e))
        except RuntimeError as e:
            raise GenericForbidden(str(e))


async def get_user_scopes(
    request: web.Request,
    params: Optional[dict[str, Any]] = None,
) -> tuple[uuid.UUID, UserRole]:
    root_ctx: RootContext = request.app["_root.context"]
    if not request["is_authorized"]:
        raise GenericForbidden("Only authorized requests may have user scopes.")
    if params is not None and (owner_user_email := params.get("owner_user_email")) is not None:
        if not request["is_superadmin"]:
            raise InvalidAPIParameters("Only superadmins may have user scopes.")
        async with root_ctx.db.begin_readonly() as conn:
            user_query = (
                sa.select([users.c.uuid, users.c.role, users.c.domain_name])
                .select_from(users)
                .where(
                    (users.c.email == owner_user_email),
                )
            )
            result = await conn.execute(user_query)
            row = result.first()
            if row is None:
                raise InvalidAPIParameters("Cannot delegate an unknown user")
            owner_user_uuid = row["uuid"]
            owner_user_role = row["role"]
            owner_user_domain = row["domain_name"]
        if request["is_superadmin"]:
            pass
        elif request["is_admin"]:
            if request["user"]["domain_name"] != owner_user_domain:
                raise GenericForbidden(
                    "Domain-admins can perform operations on behalf of "
                    "other users in the same domain only.",
                )
            if owner_user_role == UserRole.SUPERADMIN:
                raise GenericForbidden(
                    "Domain-admins cannot perform operations on behalf of super-admins.",
                )
            pass
        else:
            raise GenericForbidden(
                "Only admins can perform operations on behalf of other users.",
            )
    else:
        owner_user_uuid = request["user"]["uuid"]
        owner_user_role = request["user"]["role"]
    return owner_user_uuid, owner_user_role


P = ParamSpec("P")
TAnyResponse = TypeVar("TAnyResponse", bound=web.StreamResponse)


def check_api_params(
    checker: t.Trafaret,
    loads: Callable[[str], Any] | None = None,
    query_param_checker: t.Trafaret | None = None,
    request_examples: list[Any] | None = None,
) -> Callable[
    # We mark the arg for the validated param as Any because we cannot define a generic type of
    # Trafaret's return value.
    [Callable[Concatenate[web.Request, Any, P], Awaitable[TAnyResponse]]],
    Callable[Concatenate[web.Request, P], Awaitable[TAnyResponse]],
]:
    def wrap(handler: Callable[Concatenate[web.Request, Any, P], Awaitable[TAnyResponse]]):
        @functools.wraps(handler)
        async def wrapped(request: web.Request, *args: P.args, **kwargs: P.kwargs) -> TAnyResponse:
            orig_params: Any
            body: str = ""
            try:
                body_exists = request.can_read_body
                if body_exists and request.method not in ("GET", "HEAD"):
                    body = await request.text()
                    if request.content_type == "text/yaml":
                        orig_params = yaml.load(body, Loader=yaml.BaseLoader)
                    else:
                        orig_params = (loads or json.loads)(body)
                else:
                    orig_params = dict(request.query)
                stripped_params = orig_params.copy()
                log.debug("stripped raw params: {}", mask_sensitive_keys(stripped_params))
                checked_params = checker.check(stripped_params)
                if body_exists and query_param_checker is not None:
                    query_params = query_param_checker.check(request.query)
                    kwargs["query"] = query_params
            except (json.decoder.JSONDecodeError, yaml.YAMLError, yaml.MarkedYAMLError):
                raise InvalidAPIParameters("Malformed body")
            except t.DataError as e:
                raise InvalidAPIParameters("Input validation error", extra_data=e.as_dict())
            return await handler(request, checked_params, *args, **kwargs)

        set_handler_attr(wrapped, "request_scheme", checker)
        if request_examples:
            set_handler_attr(wrapped, "request_examples", request_examples)
        return wrapped

    return wrap


LegacyBaseRequestModel: TypeAlias = BaseRequestModel


class LegacyBaseResponseModel(BaseResponseModel):
    status: Annotated[int, Field(strict=True, exclude=True, ge=100, lt=600)] = 200


TParamModel = TypeVar("TParamModel", bound=LegacyBaseRequestModel, contravariant=True)
TQueryModel = TypeVar("TQueryModel", bound=LegacyBaseRequestModel)
TResponseModel = TypeVar("TResponseModel", bound=LegacyBaseResponseModel, covariant=True)

TPydanticResponse: TypeAlias = TResponseModel | list[TResponseModel]


class THandlerFuncWithoutParam(Protocol, Generic[P, TResponseModel]):
    def __call__(
        self,
        request: web.Request,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Awaitable[TPydanticResponse | web.StreamResponse]: ...


class THandlerFuncWithParam(Protocol, Generic[P, TParamModel, TResponseModel]):
    def __call__(
        self,
        request: web.Request,
        params: TParamModel,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Awaitable[TPydanticResponse | web.StreamResponse]: ...


def ensure_stream_response_type(
    response: LegacyBaseResponseModel
    | BaseResponseModel
    | list[TResponseModel]
    | web.StreamResponse,
) -> web.StreamResponse:
    match response:
        case LegacyBaseResponseModel(status=status):
            return web.json_response(response.model_dump(mode="json"), status=status)
        case BaseResponseModel():
            return web.json_response(response.model_dump(mode="json"))
        case list():
            return web.json_response(TypeAdapter(type(response)).dump_python(response, mode="json"))
        case web.StreamResponse():
            return response
        case _:
            raise RuntimeError(f"Unsupported response type ({type(response).__mro__})")


def pydantic_response_api_handler(
    handler: THandlerFuncWithoutParam[P, TResponseModel],
) -> Handler:
    """
    Only for API handlers which does not require request body.
    For handlers with params to consume use @pydantic_params_api_handler() or
    @check_api_params() decorator (only when request param is validated with trafaret).
    """

    @functools.wraps(handler)
    async def wrapped(
        request: web.Request, *args: P.args, **kwargs: P.kwargs
    ) -> web.StreamResponse:
        response = await handler(request, *args, **kwargs)
        return ensure_stream_response_type(response)

    return wrapped


def pydantic_params_api_handler(
    checker: type[TParamModel],
    loads: Callable[[str], Any] | None = None,
    query_param_checker: type[TQueryModel] | None = None,
) -> Callable[[THandlerFuncWithParam[P, TParamModel, TResponseModel]], Handler]:
    def wrap(
        handler: THandlerFuncWithParam[P, TParamModel, TResponseModel],
    ) -> Handler:
        @functools.wraps(handler)
        async def wrapped(
            request: web.Request, *args: P.args, **kwargs: P.kwargs
        ) -> web.StreamResponse:
            orig_params: Any
            body: str = ""
            try:
                body_exists = request.can_read_body
                if body_exists:
                    body = await request.text()
                    if request.content_type == "text/yaml":
                        orig_params = yaml.load(body, Loader=yaml.BaseLoader)
                    else:
                        orig_params = (loads or json.loads)(body)
                else:
                    orig_params = dict(request.query)
                stripped_params = orig_params.copy()
                log.debug("stripped raw params: {}", mask_sensitive_keys(stripped_params))
                checked_params = checker.model_validate(stripped_params)
                if body_exists and query_param_checker is not None:
                    query_params = query_param_checker.model_validate(request.query)
                    kwargs["query"] = query_params
            except (json.decoder.JSONDecodeError, yaml.YAMLError, yaml.MarkedYAMLError):
                raise InvalidAPIParameters("Malformed body")
            except ValidationError as ex:
                first_error = ex.errors()[0]
                # Format the first validation error as the message
                # The client may refer extra_data to access the full validation errors.
                metadata = {
                    "input": first_error["input"],
                }
                if loc := first_error["loc"]:
                    metadata["loc"] = loc[0]
                metadata_formatted_items = [
                    f"type={first_error['type']}",  # format as symbol
                    *(f"{k}={v!r}" for k, v in metadata.items()),
                ]
                msg = f"{first_error['msg']} [{', '.join(metadata_formatted_items)}]"
                # To reuse the json serialization provided by pydantic, we call ex.json() and re-parse it.
                raise InvalidAPIParameters(msg, extra_data=load_json(ex.json()))
            result = await handler(request, checked_params, *args, **kwargs)
            return ensure_stream_response_type(result)

        set_handler_attr(wrapped, "request_scheme", checker)

        return wrapped

    return wrap


_danger_words = ["password", "passwd", "secret"]


def mask_sensitive_keys(data: Mapping[str, Any]) -> Mapping[str, Any]:
    """
    Returns a new cloned mapping by masking the values of
    sensitive keys with "***" from the given mapping.
    """
    sanitized = dict()
    for k, v in data.items():
        if any((w in k.lower()) for w in _danger_words):
            sanitized[k] = "***"
        else:
            sanitized[k] = v
    return sanitized


def trim_text(value: str, maxlen: int) -> str:
    if len(value) <= maxlen:
        return value
    value = value[: maxlen - 3] + "..."
    return value


class _Infinity(numbers.Number):
    def __lt__(self, o):
        return False

    def __le__(self, o):
        return False

    def __gt__(self, o):
        return True

    def __ge__(self, o):
        return False

    def __float__(self):
        return float("inf")

    def __int__(self):
        return 0xFFFF_FFFF_FFFF_FFFF  # a practical 64-bit maximum

    def __hash__(self):
        return hash(self)


numbers.Number.register(_Infinity)
Infinity = _Infinity()


def prettify_traceback(exc):
    # Make a compact stack trace string
    with io.StringIO() as buf:
        while exc is not None:
            print(f"Exception: {exc!r}", file=buf)
            if exc.__traceback__ is None:
                print("  (no traceback available)", file=buf)
            else:
                for frame in traceback.extract_tb(exc.__traceback__):
                    short_path = _rx_sitepkg_path.sub("<sitepkg>/", frame.filename)
                    print(f"  {short_path}:{frame.lineno} ({frame.name})", file=buf)
            exc = exc.__context__
        return f"Traceback:\n{buf.getvalue()}"


def catch_unexpected(log, reraise_cancellation: bool = True, raven=None):
    def _wrap(func):
        @functools.wraps(func)
        async def _wrapped(*args, **kwargs):
            try:
                return await func(*args, **kwargs)
            except asyncio.CancelledError:
                if reraise_cancellation:
                    raise
            except Exception:
                if raven:
                    raven.captureException()
                log.exception("unexpected error!")
                raise

        return _wrapped

    return _wrap


def set_handler_attr(func, key, value):
    attrs = getattr(func, "_backend_attrs", None)
    if attrs is None:
        attrs = {}
    attrs[key] = value
    setattr(func, "_backend_attrs", attrs)


def get_handler_attr(request, key, default=None):
    # When used in the aiohttp server-side codes, we should use
    # request.match_info.hanlder instead of handler passed to the middleware
    # functions because aiohttp wraps this original handler with functools.partial
    # multiple times to implement its internal middleware processing.
    attrs = getattr(request.match_info.handler, "_backend_attrs", None)
    if attrs is not None:
        return attrs.get(key, default)
    return default


async def not_impl_stub(request: web.Request) -> web.Response:
    raise NotImplementedAPI


def deprecated_stub(msg: str) -> Callable[[web.Request], Awaitable[web.StreamResponse]]:
    async def deprecated_stub_impl(request: web.Request) -> web.Response:
        raise DeprecatedAPI(extra_msg=msg)

    return deprecated_stub_impl


def chunked(iterable, n):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


_burst_last_call: float = 0.0
_burst_times: MutableMapping[Hashable, float] = dict()
_burst_counts: MutableMapping[Hashable, int] = defaultdict(int)


async def call_non_bursty(
    key: Hashable,
    coro: Callable[[], Any],
    *,
    max_bursts: int = 64,
    max_idle: Union[int, float] = 100.0,
):
    """
    Execute a coroutine once upon max_bursts bursty invocations or max_idle
    milliseconds after bursts smaller than max_bursts.
    """
    global _burst_last_call, _burst_times, _burst_counts
    if inspect.iscoroutine(coro):
        # Coroutine objects may not be called before garbage-collected
        # as this function throttles the frequency of invocation.
        # That will generate a bogus warning by the asyncio's debug facility.
        raise TypeError("You must pass coroutine function, not coroutine object.")
    now = time.monotonic()

    if now - _burst_last_call > 3.0:
        # garbage-collect keys
        cleaned_keys = []
        for k, tick in _burst_times.items():
            if now - tick > (max_idle / 1e3):
                cleaned_keys.append(k)
        for k in cleaned_keys:
            del _burst_times[k]
            _burst_counts.pop(k, None)

    last_called = _burst_times.get(key, 0)
    _burst_times[key] = now
    _burst_last_call = now
    invoke = False

    if now - last_called > (max_idle / 1e3):
        invoke = True
        _burst_counts.pop(key, None)
    else:
        _burst_counts[key] += 1
    if _burst_counts[key] >= max_bursts:
        invoke = True
        del _burst_counts[key]

    if invoke:
        if inspect.iscoroutinefunction(coro):
            return await coro()
        else:
            return coro()


class Singleton(type):
    _instances: MutableMapping[Any, Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Undefined(metaclass=Singleton):
    pass


undefined = Undefined()



import asyncio
import logging
from collections import defaultdict
from contextlib import asynccontextmanager as actxmgr
from contextlib import suppress
from typing import AsyncIterator, Iterable, override

import aiotools
import sqlalchemy as sa
from sqlalchemy.orm import load_only, noload

from ai.backend.common.events import KernelLifecycleEventReason
from ai.backend.common.types import SessionId
from ai.backend.logging import BraceStyleAdapter

from ..api.context import RootContext
from ..models import (
    DEAD_KERNEL_STATUSES,
    DEAD_SESSION_STATUSES,
    KernelRow,
    KernelStatus,
    SessionRow,
)
from .base import DEFAULT_SWEEP_INTERVAL_SEC, AbstractSweeper

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class KernelSweeper(AbstractSweeper):
    @override
    async def sweep(self) -> None:
        query = (
            sa.select(KernelRow)
            .join(SessionRow, KernelRow.session_id == SessionRow.id)
            .where(KernelRow.status.not_in({*DEAD_KERNEL_STATUSES, KernelStatus.ERROR}))
            .where(SessionRow.status.in_(DEAD_SESSION_STATUSES))
            .options(
                noload("*"),
                load_only(
                    KernelRow.id,
                    KernelRow.session_id,
                    KernelRow.agent,
                    KernelRow.agent_addr,
                    KernelRow.container_id,
                ),
            )
        )

        async with self._db.begin_readonly() as conn:
            result = await conn.execute(query)
            kernels = result.fetchall()

        kernels_per_session = defaultdict(list)
        for kernel in kernels:
            kernels_per_session[kernel.session_id].append(kernel)

        async def _destroy_kernels(session_id: SessionId, kernels: Iterable[KernelRow]) -> None:
            try:
                await self._registry.destroy_session_lowlevel(
                    session_id,
                    [
                        {
                            "id": kernel.id,
                            "session_id": kernel.session_id,
                            "agent": kernel.agent,
                            "agent_addr": kernel.agent_addr,
                            "container_id": kernel.container_id,
                        }
                        for kernel in kernels
                    ],
                    reason=KernelLifecycleEventReason.HANG_TIMEOUT,
                )
            except Exception as e:
                self._sweeper_metric.observe_kernel_sweep(success=False)
                log.error("sweep(kernel) - failed to terminate kernels (s:{}).", session_id)
                raise e
            self._sweeper_metric.observe_kernel_sweep(success=True)
            log.info("sweep(kernel) - succeeded to terminate kernels (s:{}).", session_id)

        results_and_exceptions = await asyncio.gather(
            *[
                asyncio.create_task(_destroy_kernels(session_id, session_kernels))
                for session_id, session_kernels in kernels_per_session.items()
            ],
            return_exceptions=True,
        )
        results = [
            result_or_exception
            for result_or_exception in results_and_exceptions
            if not isinstance(result_or_exception, (BaseException, Exception))
        ]

        if kernels:
            log.info(
                "sweep(kernel) - {} orphan kernel(s) found, {} kernel(s) sweeped.",
                len(kernels),
                len(results),
            )
        else:
            log.debug("sweep(kernel) - No orphan kernels found.")


@actxmgr
async def stale_kernel_sweeper_ctx(root_ctx: RootContext) -> AsyncIterator[None]:
    async def _sweep(interval: float) -> None:
        await KernelSweeper(root_ctx.db, root_ctx.registry, root_ctx.metrics.sweeper).sweep()

    task = aiotools.create_timer(_sweep, interval=DEFAULT_SWEEP_INTERVAL_SEC)

    yield

    if not task.done():
        task.cancel()
        with suppress(asyncio.CancelledError):
            await task



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.domain.actions.base import DomainAction
from ai.backend.manager.services.domain.types import (
    DomainCreator,
    DomainData,
    UserInfo,
)


@dataclass
class CreateDomainNodeAction(DomainAction):
    user_info: UserInfo
    creator: DomainCreator
    scaling_groups: Optional[list[str]] = None

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self) -> str:
        return "create"


@dataclass
class CreateDomainNodeActionResult(BaseActionResult):
    domain_data: Optional[DomainData]
    success: bool
    description: str

    @override
    def entity_id(self) -> Optional[str]:
        return self.domain_data.name if self.domain_data is not None else None



from __future__ import annotations

import uuid
from collections.abc import Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Mapping,
    Optional,
    Self,
    Type,
    TypeVar,
    cast,
)

import graphene
import sqlalchemy as sa
from dateutil.parser import parse as dtparse
from graphene.types.datetime import DateTime as GQLDateTime
from redis.asyncio import Redis
from sqlalchemy.engine.row import Row
from sqlalchemy.orm import noload, selectinload

from ai.backend.common import msgpack, redis_helper
from ai.backend.common.types import (
    AccessKey,
    AgentId,
    BinarySize,
    KernelId,
    SessionId,
)
from ai.backend.manager.models.base import (
    Item,
    PaginatedList,
    batch_multiresult,
    batch_multiresult_in_scalar_stream,
    batch_result,
)
from ai.backend.manager.models.group import groups
from ai.backend.manager.models.image import ImageRow
from ai.backend.manager.models.minilang import JSONFieldItem
from ai.backend.manager.models.minilang.ordering import ColumnMapType, QueryOrderParser
from ai.backend.manager.models.minilang.queryfilter import (
    FieldSpecType,
    QueryFilterParser,
)

from ...defs import DEFAULT_ROLE
from ..gql_relay import AsyncNode, Connection
from ..kernel import (
    AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES,
    DEFAULT_KERNEL_ORDERING,
    LIVE_STATUS,
    KernelRow,
    KernelStatus,
    kernels,
)
from ..user import UserRole, users
from .base import (
    BigInt,
)
from .image import ImageNode

if TYPE_CHECKING:
    from ..gql import GraphQueryContext

__all__ = (
    "KernelNode",
    "KernelConnection",
    "ComputeContainer",
    "ComputeContainerList",
    "LegacyComputeSession",
    "LegacyComputeSessionList",
)


class KernelNode(graphene.ObjectType):
    class Meta:
        interfaces = (AsyncNode,)
        description = "Added in 24.09.0."

    # identity
    row_id = graphene.UUID(description="ID of kernel.")
    cluster_idx = graphene.Int()
    local_rank = graphene.Int()
    cluster_role = graphene.String()
    cluster_hostname = graphene.String()
    session_id = graphene.UUID()

    # image
    image = graphene.Field(ImageNode)
    image_reference = graphene.String(description="Added in 25.4.0.")
    architecture = graphene.String(
        description="Added in 25.4.0. The architecture that the image of this kernel requires"
    )

    # status
    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    status_data = graphene.JSONString()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()
    scheduled_at = GQLDateTime()

    # resources
    agent_id = graphene.String()
    agent_addr = graphene.String()
    container_id = graphene.String()
    resource_opts = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    live_stat = graphene.JSONString()
    abusing_report = graphene.JSONString()
    preopen_ports = graphene.List(lambda: graphene.Int)

    @classmethod
    async def batch_load_by_session_id(
        cls,
        graph_ctx: GraphQueryContext,
        session_ids: Sequence[SessionId],
    ) -> Sequence[Sequence[Self]]:
        async with graph_ctx.db.begin_readonly_session() as db_sess:
            query = sa.select(KernelRow).where(KernelRow.session_id.in_(session_ids))
            return await batch_multiresult_in_scalar_stream(
                graph_ctx,
                db_sess,
                query,
                cls,
                session_ids,
                lambda row: row.session_id,
            )

    @classmethod
    async def batch_load_by_agent_id(
        cls,
        graph_ctx: GraphQueryContext,
        agent_ids: Sequence[AgentId],
    ) -> Sequence[Sequence[Self]]:
        async with graph_ctx.db.begin_readonly_session() as db_sess:
            query = sa.select(KernelRow).where(KernelRow.agent.in_(agent_ids))
            return await batch_multiresult_in_scalar_stream(
                graph_ctx,
                db_sess,
                query,
                cls,
                agent_ids,
                lambda row: row.agent,
            )

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: KernelRow) -> Self:
        # TODO: Replace 'hide-agents' option to RBAC
        is_superadmin = ctx.user["role"] == UserRole.SUPERADMIN
        if is_superadmin:
            hide_agents = False
        else:
            hide_agents = ctx.local_config["manager"]["hide-agents"]
        status_history = row.status_history or {}
        return KernelNode(
            id=row.id,  # auto-converted to Relay global ID
            row_id=row.id,
            cluster_idx=row.cluster_idx,
            cluster_hostname=row.cluster_hostname,
            local_rank=row.local_rank,
            cluster_role=row.cluster_role,
            session_id=row.session_id,
            architecture=row.architecture,
            image_reference=row.image,
            status=row.status,
            status_changed=row.status_changed,
            status_info=row.status_info,
            status_data=row.status_data,
            created_at=row.created_at,
            terminated_at=row.terminated_at,
            starts_at=row.starts_at,
            scheduled_at=status_history.get(KernelStatus.SCHEDULED.name),
            occupied_slots=row.occupied_slots.to_json(),
            agent_id=row.agent if not hide_agents else None,
            agent_addr=row.agent_addr if not hide_agents else None,
            container_id=row.container_id if not hide_agents else None,
            resource_opts=row.resource_opts,
            preopen_ports=row.preopen_ports,
        )

    async def resolve_image(self, info: graphene.ResolveInfo) -> Optional[ImageNode]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, ImageNode.batch_load_by_name_and_arch
        )
        images = cast(list[ImageNode], await loader.load((self.image_reference, self.architecture)))
        try:
            return images[0]
        except IndexError:
            return None

    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> dict[str, Any] | None:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader_by_func(
            graph_ctx, self.batch_load_live_stat
        )
        return await loader.load(self.row_id)

    @classmethod
    async def batch_load_live_stat(
        cls, ctx: GraphQueryContext, kernel_ids: Sequence[KernelId]
    ) -> list[dict[str, Any] | None]:
        async def _pipe_builder(r: Redis):
            pipe = r.pipeline()
            for kid in kernel_ids:
                await pipe.get(str(kid))
            return pipe

        ret: list[dict[str, Any] | None] = []
        for stat in await redis_helper.execute(ctx.redis_stat, _pipe_builder):
            if stat is not None:
                ret.append(msgpack.unpackb(stat))
            else:
                ret.append(None)

        return ret


class KernelConnection(Connection):
    class Meta:
        node = KernelNode
        description = "Added in 24.09.0."


class ComputeContainer(graphene.ObjectType):
    class Meta:
        interfaces = (Item,)

    # identity
    idx = graphene.Int()  # legacy
    role = graphene.String()  # legacy
    hostname = graphene.String()  # legacy
    kernel_id = graphene.UUID(description="Added in 24.03.1.")
    cluster_idx = graphene.Int()
    local_rank = graphene.Int()
    cluster_role = graphene.String()
    cluster_hostname = graphene.String()
    session_id = graphene.UUID()  # owner session

    # image
    image = graphene.String(description="Deprecated since 24.03.0; use image_object.name")
    image_object = graphene.Field(ImageNode, description="Added in 24.03.0.")
    architecture = graphene.String()
    registry = graphene.String()

    # status
    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    status_data = graphene.JSONString()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    starts_at = GQLDateTime()
    scheduled_at = GQLDateTime()
    abusing_report = graphene.JSONString()

    # resources
    agent = graphene.String()
    agent_addr = graphene.String()
    container_id = graphene.String()
    resource_opts = graphene.JSONString()
    occupied_slots = graphene.JSONString()
    live_stat = graphene.JSONString()
    last_stat = graphene.JSONString()
    preopen_ports = graphene.List(lambda: graphene.Int, required=False)

    @classmethod
    def parse_row(cls, ctx: GraphQueryContext, row: KernelRow) -> Mapping[str, Any]:
        assert row is not None
        from .user import UserRole

        is_superadmin = ctx.user["role"] == UserRole.SUPERADMIN
        if is_superadmin:
            hide_agents = False
        else:
            hide_agents = ctx.local_config["manager"]["hide-agents"]
        status_history = row.status_history or {}
        return {
            # identity
            "id": row.id,
            "kernel_id": row.id,
            "idx": row.cluster_idx,
            "role": row.cluster_role,
            "hostname": row.cluster_hostname,
            "cluster_idx": row.cluster_idx,
            "local_rank": row.local_rank,
            "cluster_role": row.cluster_role,
            "cluster_hostname": row.cluster_hostname,
            "session_id": row.session_id,
            # image
            "image": row.image,
            "image_object": ImageNode.from_row(ctx, row.image_row),
            "architecture": row.architecture,
            "registry": row.registry,
            # status
            "status": row.status.name,
            "status_changed": row.status_changed,
            "status_info": row.status_info,
            "status_data": row.status_data,
            "created_at": row.created_at,
            "terminated_at": row.terminated_at,
            "starts_at": row.starts_at,
            "scheduled_at": status_history.get(KernelStatus.SCHEDULED.name),
            "occupied_slots": row.occupied_slots.to_json(),
            # resources
            "agent": row.agent if not hide_agents else None,
            "agent_addr": row.agent_addr if not hide_agents else None,
            "container_id": row.container_id if not hide_agents else None,
            "resource_opts": row.resource_opts,
            "preopen_ports": row.preopen_ports,
            # statistics
            # last_stat is resolved by Graphene (resolve_last_stat method)
        }

    @classmethod
    def from_row(cls, ctx: GraphQueryContext, row: Row) -> Optional[ComputeContainer]:
        if row is None:
            return None
        props = cls.parse_row(ctx, row)
        return cls(**props)

    # last_stat also fetches data from Redis, meaning that
    # both live_stat and last_stat will reference same data from same source
    # we can leave last_stat value for legacy support, as an alias to last_stat
    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(graph_ctx, "KernelStatistics.by_kernel")
        return await loader.load(self.id)

    async def resolve_last_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        return await self.resolve_live_stat(info)

    async def resolve_abusing_report(
        self,
        info: graphene.ResolveInfo,
        access_key: AccessKey | None,
    ) -> Optional[Mapping[str, Any]]:
        graph_ctx: GraphQueryContext = info.context
        if access_key is None:
            return None
        return await graph_ctx.registry.get_abusing_report(self.id)

    _queryfilter_fieldspec: FieldSpecType = {
        "image": ("image", None),
        "architecture": ("architecture", None),
        "agent": ("agent", None),
        "agent_addr": ("agent_addr", None),
        "cluster_idx": ("cluster_idx", None),
        "local_rank": ("local_rank", None),
        "cluster_role": ("cluster_role", None),
        "cluster_hostname": ("cluster_hostname", None),
        "status": ("status", KernelStatus),
        "status_info": ("status_info", None),
        "created_at": ("created_at", dtparse),
        "status_changed": ("status_changed", dtparse),
        "terminated_at": ("terminated_at", dtparse),
        "scheduled_at": (JSONFieldItem("status_history", KernelStatus.SCHEDULED.name), dtparse),
    }

    _queryorder_colmap: ColumnMapType = {
        "image": ("image", None),
        "architecture": ("architecture", None),
        "agent": ("agent", None),
        "agent_addr": ("agent_addr", None),
        "cluster_idx": ("cluster_idx", None),
        "local_rank": ("local_rank", None),
        "cluster_role": ("cluster_role", None),
        "cluster_hostname": ("cluster_hostname", None),
        "status": ("status", None),
        "status_info": ("status_info", None),
        "status_changed": ("status_changed", None),
        "created_at": ("created_at", None),
        "terminated_at": ("terminated_at", None),
        "scheduled_at": (JSONFieldItem("status_history", KernelStatus.SCHEDULED.name), None),
    }

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        session_id: SessionId,
        *,
        cluster_role: Optional[str] = None,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[str] = None,
        filter: Optional[str] = None,
    ) -> int:
        query = (
            sa.select([sa.func.count()])
            .select_from(KernelRow)
            .where(KernelRow.session_id == session_id)
        )
        if cluster_role is not None:
            query = query.where(KernelRow.cluster_role == cluster_role)
        if domain_name is not None:
            query = query.where(KernelRow.domain_name == domain_name)
        if group_id is not None:
            query = query.where(KernelRow.group_id == group_id)
        if access_key is not None:
            query = query.where(KernelRow.access_key == access_key)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        async with ctx.db.begin_readonly_session() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        session_id: SessionId,
        *,
        cluster_role: Optional[str] = None,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[AccessKey] = None,
        filter: Optional[str] = None,
        order: Optional[str] = None,
    ) -> Sequence[Optional[ComputeContainer]]:
        query = (
            sa.select(KernelRow)
            .where(KernelRow.session_id == session_id)
            .limit(limit)
            .offset(offset)
            .options(selectinload(KernelRow.image_row).options(selectinload(ImageRow.aliases)))
        )
        if cluster_role is not None:
            query = query.where(KernelRow.cluster_role == cluster_role)
        if domain_name is not None:
            query = query.where(KernelRow.domain_name == domain_name)
        if group_id is not None:
            query = query.where(KernelRow.group_id == group_id)
        if access_key is not None:
            query = query.where(KernelRow.access_key == access_key)
        if filter is not None:
            qfparser = QueryFilterParser(cls._queryfilter_fieldspec)
            query = qfparser.append_filter(query, filter)
        if order is not None:
            qoparser = QueryOrderParser(cls._queryorder_colmap)
            query = qoparser.append_ordering(query, order)
        else:
            query = query.order_by(*DEFAULT_KERNEL_ORDERING)
        async with ctx.db.begin_readonly_session() as db_session:
            return [cls.from_row(ctx, r) async for r in (await db_session.stream_scalars(query))]

    @classmethod
    async def batch_load_by_session(
        cls,
        ctx: GraphQueryContext,
        session_ids: Sequence[SessionId],
    ) -> Sequence[Sequence[ComputeContainer]]:
        query = (
            sa.select(KernelRow)
            # TODO: use "owner session ID" when we implement multi-container session
            .where(KernelRow.session_id.in_(session_ids))
            .options(selectinload(KernelRow.image_row).options(selectinload(ImageRow.aliases)))
        )
        async with ctx.db.begin_readonly_session() as conn:
            return await batch_multiresult(
                ctx,
                conn,
                query,
                cls,
                session_ids,
                lambda row: row.session_id,
            )

    @classmethod
    async def batch_load_by_agent_id(
        cls,
        ctx: GraphQueryContext,
        agent_ids: Sequence[AgentId],
        *,
        status: Optional[KernelStatus] = None,
    ) -> Sequence[Sequence[ComputeContainer]]:
        query_stmt = (
            sa.select(KernelRow)
            .where(KernelRow.agent.in_(agent_ids))
            .options(selectinload(KernelRow.image_row).options(selectinload(ImageRow.aliases)))
        )
        kernel_status: tuple[KernelStatus, ...]
        if status is not None:
            kernel_status = (status,)
        else:
            kernel_status = AGENT_RESOURCE_OCCUPYING_KERNEL_STATUSES
        query_stmt = query_stmt.where(KernelRow.status.in_(kernel_status))
        async with ctx.db.begin_readonly_session() as db_session:
            return await batch_multiresult_in_scalar_stream(
                ctx,
                db_session,
                query_stmt,
                cls,
                agent_ids,
                lambda row: row.agent,
            )

    @classmethod
    async def batch_load_detail(
        cls,
        ctx: GraphQueryContext,
        container_ids: Sequence[KernelId],
        *,
        domain_name: Optional[str] = None,
        access_key: Optional[AccessKey] = None,
    ) -> Sequence[Optional[ComputeContainer]]:
        query = (
            sa.select(KernelRow)
            .where(
                (KernelRow.id.in_(container_ids)),
            )
            .options(
                noload("*"),
                selectinload(KernelRow.group_row),
                selectinload(KernelRow.user_row),
                selectinload(KernelRow.image_row),
            )
        )
        if domain_name is not None:
            query = query.where(KernelRow.domain_name == domain_name)
        if access_key is not None:
            query = query.where(KernelRow.access_key == access_key)
        async with ctx.db.begin_readonly_session() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                container_ids,
                lambda row: row.id,
            )


class ComputeContainerList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(ComputeContainer, required=True)


# --------- pre-v5 legacy -----------

MetricValueType = TypeVar("MetricValueType", int, float)


class LegacyComputeSession(graphene.ObjectType):
    """
    Represents a main session.
    """

    class Meta:
        interfaces = (Item,)

    tag = graphene.String()  # Only for ComputeSession
    sess_id = graphene.String()  # legacy
    sess_type = graphene.String()  # legacy
    session_name = graphene.String()
    session_type = graphene.String()
    role = graphene.String()
    image = graphene.String()
    architecture = graphene.String()
    registry = graphene.String()
    domain_name = graphene.String()
    group_name = graphene.String()
    group_id = graphene.UUID()
    scaling_group = graphene.String()
    user_uuid = graphene.UUID()
    access_key = graphene.String()

    status = graphene.String()
    status_changed = GQLDateTime()
    status_info = graphene.String()
    created_at = GQLDateTime()
    terminated_at = GQLDateTime()
    startup_command = graphene.String()
    result = graphene.String()

    # hidable fields by configuration
    agent = graphene.String()
    container_id = graphene.String()

    service_ports = graphene.JSONString()

    occupied_slots = graphene.JSONString()
    occupied_shares = graphene.JSONString()
    mounts = graphene.List(lambda: graphene.List(lambda: graphene.String))
    resource_opts = graphene.JSONString()

    num_queries = BigInt()
    live_stat = graphene.JSONString()
    last_stat = graphene.JSONString()

    user_email = graphene.String()

    # Legacy fields
    lang = graphene.String()
    mem_slot = graphene.Int()
    cpu_slot = graphene.Float()
    gpu_slot = graphene.Float()
    tpu_slot = graphene.Float()
    cpu_used = BigInt()
    cpu_using = graphene.Float()
    mem_max_bytes = BigInt()
    mem_cur_bytes = BigInt()
    net_rx_bytes = BigInt()
    net_tx_bytes = BigInt()
    io_read_bytes = BigInt()
    io_write_bytes = BigInt()
    io_max_scratch_size = BigInt()
    io_cur_scratch_size = BigInt()

    # last_stat also fetches data from Redis, meaning that
    # both live_stat and last_stat will reference same data from same source
    # we can leave last_stat value for legacy support, as an alias to last_stat
    async def resolve_live_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        graph_ctx: GraphQueryContext = info.context
        loader = graph_ctx.dataloader_manager.get_loader(graph_ctx, "KernelStatistics.by_kernel")
        return await loader.load(self.id)

    async def resolve_last_stat(self, info: graphene.ResolveInfo) -> Optional[Mapping[str, Any]]:
        return await self.resolve_live_stat(info)

    async def _resolve_legacy_metric(
        self,
        info: graphene.ResolveInfo,
        metric_key: str,
        metric_field: str,
        convert_type: Type[MetricValueType],
    ) -> Optional[MetricValueType]:
        if not hasattr(self, "status"):
            return None
        graph_ctx: GraphQueryContext = info.context
        if KernelStatus[self.status] not in LIVE_STATUS:
            if self.last_stat is None:
                return convert_type(0)
            metric = self.last_stat.get(metric_key)
            if metric is None:
                return convert_type(0)
            value = metric.get(metric_field)
            if value is None:
                return convert_type(0)
            return convert_type(value)
        else:
            loader = graph_ctx.dataloader_manager.get_loader(
                graph_ctx, "KernelStatistics.by_kernel"
            )
            kstat = await loader.load(self.id)
            if kstat is None:
                return convert_type(0)
            metric = kstat.get(metric_key)
            if metric is None:
                return convert_type(0)
            value = metric.get(metric_field)
            if value is None:
                return convert_type(0)
            return convert_type(value)

    async def resolve_cpu_used(self, info: graphene.ResolveInfo) -> Optional[float]:
        return await self._resolve_legacy_metric(info, "cpu_used", "current", float)

    async def resolve_cpu_using(self, info: graphene.ResolveInfo) -> Optional[float]:
        return await self._resolve_legacy_metric(info, "cpu_util", "pct", float)

    async def resolve_mem_max_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "mem", "stats.max", int)

    async def resolve_mem_cur_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "mem", "current", int)

    async def resolve_net_rx_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "net_rx", "stats.rate", int)

    async def resolve_net_tx_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "net_tx", "stats.rate", int)

    async def resolve_io_read_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "io_read", "current", int)

    async def resolve_io_write_bytes(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "io_write", "current", int)

    async def resolve_io_max_scratch_size(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "io_scratch_size", "stats.max", int)

    async def resolve_io_cur_scratch_size(self, info: graphene.ResolveInfo) -> Optional[int]:
        return await self._resolve_legacy_metric(info, "io_scratch_size", "current", int)

    @classmethod
    def parse_row(cls, ctx: GraphQueryContext, row: Row) -> Mapping[str, Any]:
        assert row is not None
        from .user import UserRole

        mega = 2**20
        is_superadmin = ctx.user["role"] == UserRole.SUPERADMIN
        if is_superadmin:
            hide_agents = False
        else:
            hide_agents = ctx.local_config["manager"]["hide-agents"]
        return {
            "id": row["id"],
            "sess_id": row["session_name"],  # legacy, will be deprecated
            "sess_type": row["session_type"].name,  # legacy, will be deprecated
            "session_name": row["session_name"],
            "session_type": row["session_type"].name,
            "role": row["cluster_role"],
            "tag": row["tag"],
            "image": row["image"],
            "architecture": row["architecture"],
            "registry": row["registry"],
            "domain_name": row["domain_name"],
            "group_name": row[
                "name"
            ],  # group.name (group is omitted since use_labels=True is not used)
            "group_id": row["group_id"],
            "scaling_group": row["scaling_group"],
            "user_uuid": row["user_uuid"],
            "access_key": row["access_key"],
            "status": row["status"].name,
            "status_changed": row["status_changed"],
            "status_info": row["status_info"],
            "created_at": row["created_at"],
            "terminated_at": row["terminated_at"],
            "startup_command": row["startup_command"],
            "result": row["result"].name,
            "service_ports": row["service_ports"],
            "occupied_slots": row["occupied_slots"].to_json(),
            "resource_opts": row["resource_opts"],
            "num_queries": row["num_queries"],
            # optionally hidden
            "agent": row["agent"] if not hide_agents else None,
            "container_id": row["container_id"] if not hide_agents else None,
            # live_stat is resolved by Graphene
            # last_stat is resolved by Graphene
            "user_email": row["email"],
            # Legacy fields
            # NOTE: currently graphene always uses resolve methods!
            "cpu_used": 0,
            "mem_max_bytes": 0,
            "mem_cur_bytes": 0,
            "net_rx_bytes": 0,
            "net_tx_bytes": 0,
            "io_read_bytes": 0,
            "io_write_bytes": 0,
            "io_max_scratch_size": 0,
            "io_cur_scratch_size": 0,
            "lang": row["image"],
            "occupied_shares": row["occupied_shares"],
            "mem_slot": BinarySize.from_str(row["occupied_slots"].get("mem", 0)) // mega,
            "cpu_slot": float(row["occupied_slots"].get("cpu", 0)),
            "gpu_slot": float(row["occupied_slots"].get("cuda.device", 0)),
            "tpu_slot": float(row["occupied_slots"].get("tpu.device", 0)),
        }

    @classmethod
    def from_row(cls, context: GraphQueryContext, row: Row) -> Optional[LegacyComputeSession]:
        if row is None:
            return None
        props = cls.parse_row(context, row)
        return cls(**props)

    @classmethod
    async def load_count(
        cls,
        ctx: GraphQueryContext,
        *,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[AccessKey] = None,
        status: Optional[str] = None,
    ) -> int:
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(",")]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        query = (
            sa.select([sa.func.count()])
            .select_from(kernels)
            .where(kernels.c.cluster_role == DEFAULT_ROLE)
        )
        if domain_name is not None:
            query = query.where(kernels.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(kernels.c.group_id == group_id)
        if access_key is not None:
            query = query.where(kernels.c.access_key == access_key)
        if status is not None:
            query = query.where(kernels.c.status.in_(status_list))
        async with ctx.db.begin_readonly() as conn:
            result = await conn.execute(query)
            return result.scalar()

    @classmethod
    async def load_slice(
        cls,
        ctx: GraphQueryContext,
        limit: int,
        offset: int,
        *,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        access_key: Optional[AccessKey] = None,
        status: Optional[str] = None,
        order_key: Optional[str] = None,
        order_asc: bool = True,
    ) -> Sequence[LegacyComputeSession]:
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(",")]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        if order_key is None:
            _ordering = DEFAULT_KERNEL_ORDERING
        else:
            _order_func = sa.asc if order_asc else sa.desc
            _ordering = [_order_func(getattr(kernels.c, order_key))]
        j = kernels.join(groups, groups.c.id == kernels.c.group_id).join(
            users, users.c.uuid == kernels.c.user_uuid
        )
        query = (
            sa.select([kernels, groups.c.name, users.c.email])
            .select_from(j)
            .where(kernels.c.cluster_role == DEFAULT_ROLE)
            .order_by(*_ordering)
            .limit(limit)
            .offset(offset)
        )
        if domain_name is not None:
            query = query.where(kernels.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(kernels.c.group_id == group_id)
        if access_key is not None:
            query = query.where(kernels.c.access_key == access_key)
        if status is not None:
            query = query.where(kernels.c.status.in_(status_list))
        async with ctx.db.begin_readonly() as conn:
            return [
                obj
                async for r in (await conn.stream(query))
                if (obj := cls.from_row(ctx, r)) is not None
            ]

    @classmethod
    async def batch_load(
        cls,
        ctx: GraphQueryContext,
        access_keys: AccessKey,
        *,
        domain_name: Optional[str] = None,
        group_id: Optional[uuid.UUID] = None,
        status: Optional[str] = None,
    ) -> Sequence[Optional[LegacyComputeSession]]:
        j = kernels.join(groups, groups.c.id == kernels.c.group_id).join(
            users, users.c.uuid == kernels.c.user_uuid
        )
        query = (
            sa.select([kernels, groups.c.name, users.c.email])
            .select_from(j)
            .where(
                (kernels.c.access_key.in_(access_keys)) & (kernels.c.cluster_role == DEFAULT_ROLE),
            )
            .order_by(
                sa.desc(
                    sa.func.greatest(
                        kernels.c.created_at,
                        kernels.c.terminated_at,
                        kernels.c.status_changed,
                    )
                ),
            )
            .limit(100)
        )
        if domain_name is not None:
            query = query.where(kernels.c.domain_name == domain_name)
        if group_id is not None:
            query = query.where(kernels.c.group_id == group_id)
        if status is not None:
            query = query.where(kernels.c.status == status)
        async with ctx.db.begin_readonly() as conn:
            return await batch_result(
                ctx,
                conn,
                query,
                cls,
                access_keys,
                lambda row: row["access_key"],
            )

    @classmethod
    async def batch_load_detail(
        cls,
        ctx: GraphQueryContext,
        sess_ids: Sequence[SessionId],
        *,
        domain_name: Optional[str] = None,
        access_key: Optional[AccessKey] = None,
        status: Optional[str] = None,
    ) -> Sequence[Sequence[LegacyComputeSession]]:
        status_list = []
        if isinstance(status, str):
            status_list = [KernelStatus[s] for s in status.split(",")]
        elif isinstance(status, KernelStatus):
            status_list = [status]
        elif status is None:
            status_list = [KernelStatus["RUNNING"]]
        j = kernels.join(groups, groups.c.id == kernels.c.group_id).join(
            users, users.c.uuid == kernels.c.user_uuid
        )
        query = (
            sa.select([kernels, groups.c.name, users.c.email])
            .select_from(j)
            .where((kernels.c.cluster_role == DEFAULT_ROLE) & (kernels.c.session_id.in_(sess_ids)))
        )
        if domain_name is not None:
            query = query.where(kernels.c.domain_name == domain_name)
        if access_key is not None:
            query = query.where(kernels.c.access_key == access_key)
        if status_list:
            query = query.where(kernels.c.status.in_(status_list))
        async with ctx.db.begin_readonly() as conn:
            return await batch_multiresult(
                ctx,
                conn,
                query,
                cls,
                sess_ids,
                lambda row: row["session_id"],
            )


class LegacyComputeSessionList(graphene.ObjectType):
    class Meta:
        interfaces = (PaginatedList,)

    items = graphene.List(LegacyComputeSession, required=True)



"""create networks table

Revision ID: 3215bc85dc45
Revises: 428215fddea3
Create Date: 2024-08-19 08:46:39.258404

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

from ai.backend.manager.models.base import GUID, SlugType

# revision identifiers, used by Alembic.
revision = "3215bc85dc45"
down_revision = "428215fddea3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "networks",
        sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("name", sa.String(length=128), nullable=False),
        sa.Column("ref_name", sa.String(length=128), nullable=False),
        sa.Column("driver", sa.String(length=64), nullable=False),
        sa.Column(
            "options", postgresql.JSONB(astext_type=sa.Text()), server_default="{}", nullable=False
        ),
        sa.Column("project", GUID(), nullable=False),
        sa.Column("domain_name", SlugType(length=64), nullable=False),
        sa.Column(
            "created_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.Column(
            "updated_at", sa.DateTime(timezone=True), server_default=sa.text("now()"), nullable=True
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_networks")),
    )
    op.add_column(
        "sessions",
        sa.Column("network_type", sa.VARCHAR, nullable=True),
    )
    op.add_column("sessions", sa.Column("network_id", sa.String(length=128), nullable=True))
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("sessions", "network_id")
    op.drop_column("sessions", "network_type")
    op.drop_table("networks")
    # ### end Alembic commands ###



from dataclasses import dataclass
from typing import Self

from ai.backend.common.bgtask import BackgroundTaskManager
from ai.backend.common.plugin.monitor import ErrorPluginContext
from ai.backend.common.types import RedisConnectionInfo
from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.config import SharedConfig
from ai.backend.manager.idle import IdleCheckerHost
from ai.backend.manager.models.storage import StorageSessionManager
from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
from ai.backend.manager.registry import AgentRegistry
from ai.backend.manager.services.agent.processors import AgentProcessors
from ai.backend.manager.services.agent.service import AgentService
from ai.backend.manager.services.container_registry.processors import ContainerRegistryProcessors
from ai.backend.manager.services.container_registry.service import ContainerRegistryService
from ai.backend.manager.services.domain.processors import DomainProcessors
from ai.backend.manager.services.domain.service import DomainService
from ai.backend.manager.services.group.processors import GroupProcessors
from ai.backend.manager.services.group.service import GroupService
from ai.backend.manager.services.image.processors import ImageProcessors
from ai.backend.manager.services.image.service import ImageService
from ai.backend.manager.services.keypair_resource_policy.processors import (
    KeypairResourcePolicyProcessors,
)
from ai.backend.manager.services.keypair_resource_policy.service import KeypairResourcePolicyService
from ai.backend.manager.services.metric.processors.utilization_metric import (
    UtilizationMetricProcessors,
)
from ai.backend.manager.services.metric.root_service import UtilizationMetricService
from ai.backend.manager.services.project_resource_policy.processors import (
    ProjectResourcePolicyProcessors,
)
from ai.backend.manager.services.project_resource_policy.service import ProjectResourcePolicyService
from ai.backend.manager.services.resource_preset.processors import ResourcePresetProcessors
from ai.backend.manager.services.resource_preset.service import ResourcePresetService
from ai.backend.manager.services.session.processors import SessionProcessors
from ai.backend.manager.services.session.service import SessionService, SessionServiceArgs
from ai.backend.manager.services.user.processors import UserProcessors
from ai.backend.manager.services.user.service import UserService
from ai.backend.manager.services.user_resource_policy.processors import UserResourcePolicyProcessors
from ai.backend.manager.services.user_resource_policy.service import UserResourcePolicyService
from ai.backend.manager.services.vfolder.processors import (
    VFolderFileProcessors,
    VFolderInviteProcessors,
    VFolderProcessors,
)
from ai.backend.manager.services.vfolder.services.file import VFolderFileService
from ai.backend.manager.services.vfolder.services.invite import VFolderInviteService
from ai.backend.manager.services.vfolder.services.vfolder import VFolderService


@dataclass
class ServiceArgs:
    db: ExtendedAsyncSAEngine
    shared_config: SharedConfig
    storage_manager: StorageSessionManager
    redis_stat: RedisConnectionInfo
    background_task_manager: BackgroundTaskManager
    agent_registry: AgentRegistry
    error_monitor: ErrorPluginContext
    idle_checker_host: IdleCheckerHost


@dataclass
class Services:
    agent: AgentService
    domain: DomainService
    group: GroupService
    user: UserService
    image: ImageService
    container_registry: ContainerRegistryService
    vfolder: VFolderService
    vfolder_file: VFolderFileService
    vfolder_invite: VFolderInviteService
    session: SessionService
    keypair_resource_policy: KeypairResourcePolicyService
    user_resource_policy: UserResourcePolicyService
    project_resource_policy: ProjectResourcePolicyService
    resource_preset: ResourcePresetService
    utilization_metric: UtilizationMetricService

    @classmethod
    def create(cls, args: ServiceArgs) -> Self:
        agent_service = AgentService(
            args.db,
            args.agent_registry,
            args.shared_config,
        )
        domain_service = DomainService(args.db)
        group_service = GroupService(
            args.db, args.storage_manager, args.shared_config, args.redis_stat
        )
        user_service = UserService(args.db, args.storage_manager, args.redis_stat)
        image_service = ImageService(args.db, args.agent_registry)
        container_registry_service = ContainerRegistryService(args.db)
        vfolder_service = VFolderService(
            args.db, args.shared_config, args.storage_manager, args.background_task_manager
        )
        vfolder_file_service = VFolderFileService(args.db, args.shared_config, args.storage_manager)
        vfolder_invite_service = VFolderInviteService(args.db, args.shared_config)
        session_service = SessionService(
            SessionServiceArgs(
                db=args.db,
                agent_registry=args.agent_registry,
                background_task_manager=args.background_task_manager,
                error_monitor=args.error_monitor,
                idle_checker_host=args.idle_checker_host,
            )
        )
        keypair_resource_policy_service = KeypairResourcePolicyService(args.db)
        user_resource_policy_service = UserResourcePolicyService(args.db)
        project_resource_policy_service = ProjectResourcePolicyService(args.db)
        resource_preset_service = ResourcePresetService(
            args.db, args.agent_registry, args.shared_config
        )
        utilization_metric_service = UtilizationMetricService(args.shared_config)

        return cls(
            agent=agent_service,
            domain=domain_service,
            group=group_service,
            user=user_service,
            image=image_service,
            container_registry=container_registry_service,
            vfolder=vfolder_service,
            vfolder_file=vfolder_file_service,
            vfolder_invite=vfolder_invite_service,
            session=session_service,
            keypair_resource_policy=keypair_resource_policy_service,
            user_resource_policy=user_resource_policy_service,
            project_resource_policy=project_resource_policy_service,
            resource_preset=resource_preset_service,
            utilization_metric=utilization_metric_service,
        )


@dataclass
class ProcessorArgs:
    service_args: ServiceArgs


@dataclass
class Processors:
    agent: AgentProcessors
    domain: DomainProcessors
    group: GroupProcessors
    user: UserProcessors
    image: ImageProcessors
    vfolder: VFolderProcessors
    vfolder_invite: VFolderInviteProcessors
    vfolder_file: VFolderFileProcessors
    session: SessionProcessors
    container_registry: ContainerRegistryProcessors
    keypair_resource_policy: KeypairResourcePolicyProcessors
    user_resource_policy: UserResourcePolicyProcessors
    project_resource_policy: ProjectResourcePolicyProcessors
    resource_preset: ResourcePresetProcessors
    utilization_metric: UtilizationMetricProcessors

    @classmethod
    def create(cls, args: ProcessorArgs, action_monitors: list[ActionMonitor]) -> Self:
        services = Services.create(args.service_args)
        agent_processors = AgentProcessors(services.agent, action_monitors)
        domain_processors = DomainProcessors(services.domain, action_monitors)
        group_processors = GroupProcessors(services.group, action_monitors)
        user_processors = UserProcessors(services.user, action_monitors)
        image_processors = ImageProcessors(services.image, action_monitors)
        container_registry_processors = ContainerRegistryProcessors(
            services.container_registry, action_monitors
        )
        vfolder_processors = VFolderProcessors(services.vfolder, action_monitors)
        vfolder_file_processors = VFolderFileProcessors(services.vfolder_file, action_monitors)
        vfolder_invite_processors = VFolderInviteProcessors(
            services.vfolder_invite, action_monitors
        )
        session_processors = SessionProcessors(services.session, action_monitors)
        keypair_resource_policy_processors = KeypairResourcePolicyProcessors(
            services.keypair_resource_policy, action_monitors
        )
        user_resource_policy_processors = UserResourcePolicyProcessors(
            services.user_resource_policy, action_monitors
        )
        project_resource_policy_processors = ProjectResourcePolicyProcessors(
            services.project_resource_policy, action_monitors
        )
        resource_preset_processors = ResourcePresetProcessors(
            services.resource_preset, action_monitors
        )
        utilization_metric_processors = UtilizationMetricProcessors(
            services.utilization_metric, action_monitors
        )
        return cls(
            agent=agent_processors,
            domain=domain_processors,
            group=group_processors,
            user=user_processors,
            image=image_processors,
            container_registry=container_registry_processors,
            vfolder=vfolder_processors,
            vfolder_file=vfolder_file_processors,
            vfolder_invite=vfolder_invite_processors,
            session=session_processors,
            keypair_resource_policy=keypair_resource_policy_processors,
            user_resource_policy=user_resource_policy_processors,
            project_resource_policy=project_resource_policy_processors,
            resource_preset=resource_preset_processors,
            utilization_metric=utilization_metric_processors,
        )



from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.services.group.actions.create_group import (
    CreateGroupAction,
    CreateGroupActionResult,
)
from ai.backend.manager.services.group.actions.delete_group import (
    DeleteGroupAction,
    DeleteGroupActionResult,
)
from ai.backend.manager.services.group.actions.modify_group import (
    ModifyGroupAction,
    ModifyGroupActionResult,
)
from ai.backend.manager.services.group.actions.purge_group import (
    PurgeGroupAction,
    PurgeGroupActionResult,
)
from ai.backend.manager.services.group.actions.usage_per_month import (
    UsagePerMonthAction,
    UsagePerMonthActionResult,
)
from ai.backend.manager.services.group.actions.usage_per_period import (
    UsagePerPeriodAction,
    UsagePerPeriodActionResult,
)
from ai.backend.manager.services.group.service import GroupService


class GroupProcessors:
    create_group: ActionProcessor[CreateGroupAction, CreateGroupActionResult]
    modify_group: ActionProcessor[ModifyGroupAction, ModifyGroupActionResult]
    delete_group: ActionProcessor[DeleteGroupAction, DeleteGroupActionResult]
    purge_group: ActionProcessor[PurgeGroupAction, PurgeGroupActionResult]
    usage_per_month: ActionProcessor[UsagePerMonthAction, UsagePerMonthActionResult]
    usage_per_period: ActionProcessor[UsagePerPeriodAction, UsagePerPeriodActionResult]

    def __init__(self, group_service: GroupService, action_monitors: list[ActionMonitor]) -> None:
        self.create_group = ActionProcessor(group_service.create_group, action_monitors)
        self.modify_group = ActionProcessor(group_service.modify_group, action_monitors)
        self.delete_group = ActionProcessor(group_service.delete_group, action_monitors)
        self.purge_group = ActionProcessor(group_service.purge_group, action_monitors)
        self.usage_per_month = ActionProcessor(group_service.usage_per_month, action_monitors)
        self.usage_per_period = ActionProcessor(group_service.usage_per_period, action_monitors)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.resource.types import UserResourcePolicyData
from ai.backend.manager.services.user_resource_policy.actions.base import UserResourcePolicyAction


@dataclass
class DeleteUserResourcePolicyAction(UserResourcePolicyAction):
    name: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "delete"


@dataclass
class DeleteUserResourcePolicyActionResult(BaseActionResult):
    user_resource_policy: UserResourcePolicyData

    @override
    def entity_id(self) -> Optional[str]:
        return self.user_resource_policy.name



from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseAction


@dataclass
class DomainAction(BaseAction):
    @override
    def entity_type(self):
        return "domain"



from dataclasses import dataclass, field
from typing import Any, Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.exceptions import BaseActionException
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.models.image import ImageType
from ai.backend.manager.services.image.actions.base import ImageAction
from ai.backend.manager.types import OptionalState, PartialModifier, TriState


@dataclass
class ImageModifier(PartialModifier):
    name: OptionalState[str] = field(default_factory=OptionalState.nop)
    registry: OptionalState[str] = field(default_factory=OptionalState.nop)
    image: OptionalState[str] = field(default_factory=OptionalState.nop)
    tag: OptionalState[str] = field(default_factory=OptionalState.nop)
    architecture: OptionalState[str] = field(default_factory=OptionalState.nop)
    is_local: OptionalState[bool] = field(default_factory=OptionalState.nop)
    size_bytes: OptionalState[int] = field(default_factory=OptionalState.nop)
    type: OptionalState[ImageType] = field(default_factory=OptionalState.nop)
    config_digest: OptionalState[str] = field(default_factory=OptionalState.nop)
    labels: OptionalState[dict[str, Any]] = field(default_factory=OptionalState.nop)
    accelerators: TriState[str] = field(default_factory=TriState.nop)
    resources: OptionalState[dict[str, Any]] = field(default_factory=OptionalState.nop)

    @override
    def fields_to_update(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.registry.update_dict(to_update, "registry")
        self.image.update_dict(to_update, "image")
        self.tag.update_dict(to_update, "tag")
        self.architecture.update_dict(to_update, "architecture")
        self.is_local.update_dict(to_update, "is_local")
        self.size_bytes.update_dict(to_update, "size_bytes")
        self.type.update_dict(to_update, "type")
        self.config_digest.update_dict(to_update, "config_digest")
        self.labels.update_dict(to_update, "labels")
        self.accelerators.update_dict(to_update, "accelerators")
        self.resources.update_dict(to_update, "_resources")
        return to_update


@dataclass
class ModifyImageAction(ImageAction):
    target: str
    architecture: str
    modifier: ImageModifier

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    def operation_type(self):
        return "modify"


@dataclass
class ModifyImageActionResult(BaseActionResult):
    image: ImageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image.id)


class ModifyImageActionUnknownImageReferenceError(BaseActionException):
    pass


class ModifyImageActionValueError(BaseActionException):
    pass



"""add endpoint and routing tables

Revision ID: f108628f032b
Revises: b5be363ab05c
Create Date: 2023-02-17 05:47:48.060365

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID, EndpointIDColumnType, ResourceSlotColumn

# revision identifiers, used by Alembic.
revision = "f108628f032b"
down_revision = "b5be363ab05c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "endpoints",
        sa.Column(
            "id",
            EndpointIDColumnType(),
            server_default=sa.text("uuid_generate_v4()"),
            nullable=False,
        ),
        sa.Column("image", GUID(), nullable=False),
        sa.Column("model", GUID(), nullable=False),
        sa.Column("domain", sa.String(length=64), nullable=False),
        sa.Column("project", GUID(), nullable=False),
        sa.Column("resource_group", sa.String(length=64), nullable=False),
        sa.Column("resource_slots", ResourceSlotColumn(astext_type=sa.Text()), nullable=False),
        sa.Column("url", sa.String(length=1024), nullable=False),
        sa.ForeignKeyConstraint(
            ["domain"],
            ["domains.name"],
            name=op.f("fk_endpoints_domain_domains"),
            ondelete="RESTRICT",
        ),
        sa.ForeignKeyConstraint(
            ["image"], ["images.id"], name=op.f("fk_endpoints_image_images"), ondelete="RESTRICT"
        ),
        sa.ForeignKeyConstraint(
            ["model"],
            ["vfolders.id"],
            name=op.f("fk_endpoints_model_vfolders"),
            ondelete="RESTRICT",
        ),
        sa.ForeignKeyConstraint(
            ["project"],
            ["groups.id"],
            name=op.f("fk_endpoints_project_groups"),
            ondelete="RESTRICT",
        ),
        sa.ForeignKeyConstraint(
            ["resource_group"],
            ["scaling_groups.name"],
            name=op.f("fk_endpoints_resource_group_scaling_groups"),
            ondelete="RESTRICT",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_endpoints")),
        sa.UniqueConstraint("url", name=op.f("uq_endpoints_url")),
    )
    op.create_index(
        op.f("ix_endpoints_resource_group"), "endpoints", ["resource_group"], unique=False
    )
    op.create_table(
        "routings",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("endpoint", GUID(), nullable=False),
        sa.Column("session", GUID(), nullable=False),
        sa.Column("traffic_ratio", sa.Float(), nullable=False),
        sa.ForeignKeyConstraint(
            ["endpoint"],
            ["endpoints.id"],
            name=op.f("fk_routings_endpoint_endpoints"),
            ondelete="CASCADE",
        ),
        sa.ForeignKeyConstraint(
            ["session"],
            ["sessions.id"],
            name=op.f("fk_routings_session_sessions"),
            ondelete="RESTRICT",
        ),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_routings")),
        sa.UniqueConstraint("endpoint", "session", name="uq_routings_endpoint_session"),
    )
    op.create_index(op.f("ix_sessions_status"), "sessions", ["status"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_sessions_status"), table_name="sessions")
    op.drop_table("routings")
    op.drop_index(op.f("ix_endpoints_resource_group"), table_name="endpoints")
    op.drop_table("endpoints")
    # ### end Alembic commands ###



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Self,
)
from uuid import UUID

import graphene

from ai.backend.manager.services.metric.actions.container import (
    ContainerMetricAction,
)
from ai.backend.manager.services.metric.types import (
    ContainerMetricOptionalLabel,
    MetricQueryParameter,
)

from .base import ContainerUtilizationMetric, MetricResultValue

if TYPE_CHECKING:
    from ...gql import GraphQueryContext


class UserUtilizationMetricQueryInput(graphene.InputObjectType):
    class Meta:
        description = "Added in 25.6.0."

    value_type = graphene.String(
        default_value="current",
        description="One of 'current', 'capacity'. Default value is 'current'.",
    )
    metric_name = graphene.String(
        required=True,
        description="metric name of container utilization. For example, 'cpu_util', 'mem'.",
    )
    start = graphene.String(required=True, description="rfc3339 or unix_timestamp.")
    end = graphene.String(required=True, description="rfc3339 or unix_timestamp.")
    step = graphene.String(
        required=True,
        description=(
            "Query resolution step width in duration format or float number of seconds. "
            "For example, '1m', '1h', '1d', '1w'"
        ),
    )

    def metric_query_param(self) -> MetricQueryParameter:
        return MetricQueryParameter(
            metric_name=self.metric_name,
            value_type=self.value_type,
            start=self.start,
            end=self.end,
            step=self.step,
        )


class UserUtilizationMetric(graphene.ObjectType):
    class Meta:
        description = "Added in 25.6.0."

    user_id = graphene.UUID()
    metrics = graphene.List(ContainerUtilizationMetric)

    @classmethod
    async def get_object(
        cls,
        info: graphene.ResolveInfo,
        user_id: UUID,
        param: MetricQueryParameter,
    ) -> Self:
        graph_ctx: GraphQueryContext = info.context
        action_result = (
            await graph_ctx.processors.utilization_metric.query_container.wait_for_complete(
                ContainerMetricAction(
                    metric_name=param.metric_name,
                    labels=ContainerMetricOptionalLabel(
                        user_id=user_id, value_type=param.value_type
                    ),
                    start=param.start,
                    end=param.end,
                    step=param.step,
                )
            )
        )
        metrics = []
        for result in action_result.result:
            metrics.append(
                ContainerUtilizationMetric(
                    metric_name=param.metric_name,
                    value_type=result.metric.value_type,
                    values=[
                        MetricResultValue(
                            timestamp=value.timestamp,
                            value=value.value,
                        )
                        for value in result.values
                    ],
                )
            )

        return cls(
            user_id=user_id,
            metrics=metrics,
        )
